#
# Copyright 2018 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Best-practice performance tuning of bcache devices on SD/NVMe based installations."""
import os
import shutil
import subprocess

from charmhelpers.core.hookenv import (
    config,
    log,
    status_set,
)

from charms.reactive import (
    hook,
    set_flag,
    when,
)

import yaml

TUNING_CONFIG_FILE = "/etc/tune-bcache-config.yaml"
TUNING_MAP = {
    "cache_set": [
        "congested_read_threshold_us",
        "congested_write_threshold_us",
    ],
    "backing_device": [
        "readahead",
        "sequential_cutoff",
        "writeback_percent",
        "cache_mode",
    ],
}


def write_tuning_config():
    """Persist configuration parameters to configuration file."""
    conf = config()
    if conf["cache_mode"] not in [
        "writethrough",
        "writeback",
        "writearound",
        "none",
        "unmanaged",
    ]:
        log(
            "illegal value for cache_mode ({})".format(conf["cache_mode"]),
            level="ERROR",
        )
        conf["cache_mode"] = "unmanaged"

    try:
        with open(TUNING_CONFIG_FILE, "r") as f:
            tunings = yaml.safe_load(f)
    except FileNotFoundError:
        tunings = {}

    for section in TUNING_MAP.keys():
        if section not in tunings:
            tunings[section] = {}
        for option in TUNING_MAP[section]:
            tunings[section][option] = conf[option]

    with open(TUNING_CONFIG_FILE, "w") as f:
        yaml.safe_dump(tunings, f)


@hook("upgrade-charm")
def upgrade_bcache_tuning():
    """Configure a systemd service to persist configurations across reboots."""
    shutil.copyfile("files/tune-bcache", "/usr/sbin/tune-bcache")
    subprocess.check_call(
        [
            "sed",
            "-e",
            "s:TUNING_CONFIG_FILE:{}:".format(TUNING_CONFIG_FILE),
            "-i",
            "/usr/sbin/tune-bcache",
        ]
    )
    os.chmod("/usr/sbin/tune-bcache", 0o755)
    shutil.copyfile(
        "files/tune-bcache.service", "/lib/systemd/system/tune-bcache.service"
    )
    write_tuning_config()
    subprocess.check_call(["systemctl", "daemon-reload"])
    subprocess.check_call(["systemctl", "enable", "tune-bcache.service"])
    subprocess.check_call(["systemctl", "start", "tune-bcache.service"])


@hook("install")
def install_bcache_tuning():
    """Install charm."""
    upgrade_bcache_tuning()
    set_flag("bcache-tuning.installed")


@hook("stop")
def remove_bcache_tuning():
    """Remove bcahe-tuning from unit."""
    subprocess.check_call(["systemctl", "stop", "tune-bcache.service"])
    subprocess.check_call(["systemctl", "disable", "tune-bcache.service"])
    os.unlink("/lib/systemd/system/tune-bcache.service")
    os.unlink("/usr/sbin/tune-bcache")
    os.unlink(TUNING_CONFIG_FILE)
    subprocess.check_call(["systemctl", "daemon-reload"])


@when("bcache-tuning.installed")
def assess_status():
    """Keep Juju description up to date."""
    status_set("active", "bcache devices tuned")


@when("config.changed")
def update_bcache_tuning():
    """Update configuration parameters and restart systemd service."""
    write_tuning_config()
    subprocess.check_call(["systemctl", "restart", "tune-bcache.service"])
